% LIBGPUPERFCNT(8)
% Marius Vlad <marius-cristian.vlad@nxp.com>
% June 4, 2018

# NAME

**libgpuperfcnt** --  Library to read various information values from the GPU 
driver.

# SYNOPSIS

**libgpuperfcnt** can be used to abstract communication between your application
and the underlining hardware. It is currently used by **gputop** to:

* read hw and driver information
* read DMA engine status and load
* read occupancy module states
* counter values
* clients attached to the GPU
* etc

# DESCRIPTION

You can use **libgpuperfcnt** and provide your own read/write methods to
communicate with the driver. Each driver has to implement the following data structure:

	struct perf_ops {
		int (*open)(uint32_t flags, struct perf_device *dev);
		int (*open_with_fd)(int fd, uint32_t flags, struct perf_device *dev);
		int (*close)(struct perf_device *dev);

		int (*enable)(struct perf_device *dev);
		int (*enable_with_ctx)(uint32_t ctx, struct perf_device *dev);
		int (*disable)(struct perf_device *dev);

		int (*start)(struct perf_device *dev);
		int (*stop)(struct perf_device *dev);

		int (*read_counters_3d)(uint32_t type, void *data, struct perf_device *dev);
		int (*read_counters_2d)(void *data, struct perf_device *dev);

		int (*read_register)(uint32_t core_id, uint32_t reg, uint32_t *data, struct perf_device *dev);
		int (*write_register)(uint32_t core_id, uint32_t reg, uint32_t data, struct perf_device *dev);

		uint32_t (*get_num_counters)(uint32_t type);

		struct perf_counter_info *(*get_counter_info)(uint32_t type, uint32_t id);

		int (*reset_counters)(bool val, struct perf_device *dev);

		int (*profiler_is_available)(int *state);

		uint32_t *(*get_current_contexts)(uint32_t *ctx_no, struct perf_device *dev);
		int (*get_contexts)(uint32_t **ctxs_no, uint32_t *ctx_no, uint32_t pid, struct perf_device *dev);
		int (*context_set)(uint32_t ctx, struct perf_device *dev);

		int (*get_driver_info)(struct perf_driver_info *drv_info, struct perf_device *dev);

		int (*get_hw_info)(struct perf_hw_info *hw_info, struct perf_device *dev);
		int (*free_hw_info)(struct perf_hw_info *hw_info);

		int (*get_client_memory)(struct perf_client_memory *c_mem, int pid, struct perf_device *dev);

		enum perf_core_type (*get_core_type)(uint32_t core_id, struct perf_device *dev);

		const char *(*get_last_error)(struct perf_device *dev);
	};

For example, for VSI this structure is called vivante_ops.

To initialize this library for VSI the following examples (see **EXAMPLES**
section) can be used. Here &vivante_ops has already beed declared as
external in gpuperfcnt/gpuperfcnt_vivante.h header.

# EXAMPLES

## Basic initialization

	/* include library headers */
	#include <gpuperfcnt/gpuperfcnt.h>

	/* specific driver */
	#include <gpuperfcnt/gpuperfcnt_vivante.h>

	/* this holds all the info we need */
	struct perf_device *dev = NULL;

	/* initialize the perf_device with vivante_ops found in gpuperfcnt_vivante.h */
	dev = perf_init(&vivante_ops);
	if (!dev) {
		fprintf(stderr, "Failed to init perf\n");
		exit(EXIT_FAILURE);
	}

	/* open driver connection */
	if (perf_open(VIV_HW_3D, dev) < 0) {
		/* might want to check for driver version, other-wise abort */
		fprintf(stderr, "Failed to open driver connection: %s\n",
				perf_get_last_error(dev));
		goto out_err;
	}

	/* perform operations here: read counters, registers, hw_info, dma engines etc */

	/* de-allocate resources and close connection to driver */
	perf_exit(dev);

## Retrieving HW information

	struct perf_device *dev;

	struct perf_hw_info hw_info;
	struct perf_driver_info drv_info;

	/* init call backs to VSI/Vivante */
	dev = perf_init(&vivante_ops);
	if (!dev) {
		fprintf(stderr, "Failed to init perf\n");
		exit(EXIT_FAILURE);
	}

	/* open driver connection */
	if (perf_open(VIV_HW_3D, dev) < 0) {
		fprintf(stderr, "Failed to open driver connection\n");
		exit(EXIT_FAILURE);
	}

	/* retrieve driver info */
	perf_get_driver_info(&drv_info, dev);

	fprintf(stdout, "Galcore version:%d.%d.%d.%dn",
			drv_info.major, drv_info.minor,
			drv_info.patch, drv_info.build);

	/* retrive hw info */
	perf_get_hw_info(&hw_info, dev);
	for (struct perf_hw_info *hi = hw_info.head; hi != NULL; hi = hi->next) {
		fprintf(stdout, "GC %x, Rev:%x\n", hi->model, hi->revision);
	}

	/* do not leak hw info */
	perf_free_hw_info(&hw_info, dev);

	/* de-allocate resources and close connection to driver */
	perf_exit(dev);
	return 0;

## Reading Registers

	static int
	registers_show(struct registers *regs)
	{
		double cycles_idle_percent;
		double cycles_percent;

		cycles_percent = 100.0f * (double) regs->total_cycles / (double) samples;
		cycles_idle_percent = 100.0f * (double) regs->total_idle_cycles / (double) samples;

		fprintf(stdout, "RAW CYCLES: %u (0x%x)\n", regs->cycles, regs->cycles);
		fprintf(stdout, "RAW IDLE CYCLES: %u (0x%x)\n", regs->idle_cycles, regs->idle_cycles);

		fprintf(stdout, "CYCLES: %.2f%%, IDLE: %.2f%%\n", cycles_percent, cycles_idle_percent);


		return 0;
	}

	static int
	registers_read(struct registers *regs, struct perf_device *dev)
	{
		perf_read_register(PERF_MGPU_3D_CORE_0, GC_TOTAL_CYCLES, &regs->cycles, dev);
		perf_read_register(PERF_MGPU_3D_CORE_0, GC_TOTAL_IDLE_CYCLES, &regs->idle_cycles, dev);

		if (regs->cycles)
			regs->total_cycles++;

		if (regs->idle_cycles)
			regs->total_idle_cycles++;

		/* reset regs */
		perf_write_register(PERF_MGPU_3D_CORE_0, GC_TOTAL_CYCLES, 0, dev);
		perf_write_register(PERF_MGPU_3D_CORE_0, GC_TOTAL_IDLE_CYCLES, 0, dev);

		/* 1MIL ns = 100us = 1ms */
		nanosleep(&(struct timespec) { .tv_sec = 0, .tv_nsec = 1000 * 1000 }, NULL);

		return 0;
	}

	static void
	sigint_handler(int sig, siginfo_t *si, void *unused)
	{
		(void) sig;
		(void) si;
		(void) unused;

		sig_recv  = 1;
	}

	int main(void)
	{
		struct registers regs = {};
		struct perf_device *dev = NULL;
		int s;
		int state = 0;

		struct sigaction sa;
		sa.sa_flags = SA_SIGINFO;
		sigemptyset(&sa.sa_mask);

		sa.sa_sigaction = sigint_handler;

		if (sigaction(SIGTERM, &sa, NULL) == -1)
			exit(EXIT_FAILURE);
		if (sigaction(SIGINT, &sa, NULL) == -1)
			exit(EXIT_FAILURE);

		dev = perf_init(&vivante_ops);
		if (!dev) {
			fprintf(stderr, "Failed to init perf\n");
			exit(EXIT_FAILURE);
		}

		/* open driver connection */
		if (perf_open(VIV_HW_3D, dev) < 0) {
			fprintf(stderr, "Failed to open driver connection\n");
			perf_exit(dev);
			exit(EXIT_FAILURE);
		}

		while (1) {
			fprintf(stdout, "%s", clear_data_screen);

			if (sig_recv) {
				goto out;
			}

			/* reset regs entirely */
			memset(&regs, 0, sizeof(struct registers));
			for (s = 0; s < samples; s++) {
				registers_read(&regs);
			}

			registers_show(&regs);
		}
	out:
		perf_profiler_disable(dev);
		perf_exit(dev);

		return 0;
	}


# REQUIREMENTS 

This library must mach the driver version. Failure to do so would result in
perf_ops::perf_open to give an error.

# NOTES

* i.MX8QM (GC7000) variants are known to have issues while reading counters and
GPU driver will hang upon reading continously the counter values.

# SEE ALSO

*gputop(8)*
